/**
 * CCCamp07 ScoringSystem
 * A CTF scoring bot & flag+advisory reporting system
 *
 * (C) 2007, Hans-Christian Esperer
 * hc at hcespererorg
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above
 *   copyright notice, this list of conditions and the following
 *   disclaimer in the documentation and/or other materials provided
 *   with the distribution.
 * * Neither the name of the H. Ch. Esperer nor the names of his
 *   contributors may be used to endorse or promote products derived
 *   from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * POSSIBILITY OF SUCH DAMAGE
 **************************************************************************/

package de.sectud.ctf07.scoringsystem;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PrintStream;
import java.net.Socket;
import java.net.SocketException;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Calendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Random;

import net.sourceforge.adela.exceptions.FunctionNotFoundException;
import net.sourceforge.adela.exceptions.WrongParameterCountException;
import net.sourceforge.adela.interfaces.IFunctionCallback;
import net.sourceforge.adela.interpreter.ADELAInterpreter;
import net.sourceforge.adela.interpreter.ADELAParseTree;

public class ClientHandler extends Thread implements IFunctionCallback {

	private Socket socket;

	private boolean dontQuit = true;

	private PrintStream writer = System.err;

	private LimitedBufferedReader reader = null;

	private Connection connection;

	private boolean isAdmin = false;

	private int adminTriesLeft = 3;

	private long creationTime;

	private InputStream input;

	private OutputStream output;

	private byte[] ipAddress;

	private int reportedFlags = 0;

	private int calledFunctions = 0;

	private boolean ownsConnection;

	private ADELAInterpreter evaluator;

	private HashMap<String, Object> locals;

	public ClientHandler(Socket clientSocket) {
		this.socket = clientSocket;
		creationTime = Calendar.getInstance().getTimeInMillis() / 1000;
		ipAddress = clientSocket.getInetAddress().getAddress();
	}

	public void doConnect() throws FileNotFoundException,
			ClassNotFoundException, SQLException, IOException {
		this.connection = SQLConnection.getInstance().openConnection();
		this.ownsConnection = SQLConnection.getInstance()
				.usingMultipleConnections();
	}

	public long getCreationTime() {
		return creationTime;
	}

	@Override
	public void run() {
		try {
			this.socket.setSoTimeout(60000);
		} catch (SocketException e2) {
			e2.printStackTrace();
		}
		locals = new HashMap<String, Object>();
		this.evaluator = new ADELAInterpreter();
		evaluator.SetVariable("quit", "Enter quit() to quit!");
		evaluator.SetVariable("help", Manual.getInstance());
		evaluator.SetVariable("man", Manual.getInstance());
		try {
			evaluator.SetVariable("license", Manual.getInstance().getPage(
					"LICENSE"));
		} catch (IOException e1) {
			evaluator.SetVariable("license",
					"Error while reading license file!");
		}
		evaluator.AddBuiltinHandlers();
		evaluator.DeclareClassSafe(TeamHandler.class);
		evaluator.DeclareClassSafe(ServiceManager.class);
		evaluator.SetCallback(this);
		try {
			input = socket.getInputStream();
			output = socket.getOutputStream();

			reader = new LimitedBufferedReader(new InputStreamReader(input));
			writer = new PrintStream(output);
			try {
				writer.println("Initializing...");
				doConnect();
			} catch (Exception e) {
				e.printStackTrace(writer);
				writer.println("Error; giving up");
				return;
			}

			String txtHelp = "Welcome to the CTF'07 scorebot\n"
					+ "(C) 2007, Christian Esperer\n"
					+ "-----------------------------------------------------------------\n"
					+ "man\t\thelp\n" + "quit()\t\tquit\n"
					+ "dir()\t\tlist of defined variables\n"
					+ "license\t\tread the license\n";
			evaluator.SetVariable("help", txtHelp);
			writer.println(txtHelp);
			int exceptions = 0;
			int iterations = 0;
			Object lastRetVal = null;
			while (dontQuit) {
				try {
					if (iterations++ > 250) {
						dontQuit = false;
						System.out
								.println("Overflow; terminating client connection");
					}
					String rvDpl = "";
					if (lastRetVal != null) {
						rvDpl = lastRetVal.toString().split("\n")[0];
					}
					rvDpl = rvDpl.length() == 0 ? ""
							: (rvDpl.length() < 25 ? rvDpl + " " : rvDpl
									.substring(0, 22)
									+ "... ");
					writer.print(Stuff.beginPrompt() + "scorebot " + rvDpl
							+ (this.isAdmin ? "#" : ">") + Stuff.endPromph()
							+ " ");
					writer.flush();
					String line = reader.readLine().trim();
					if (line.length() != 0) {
						writer.println();
						ADELAParseTree tree = ADELAParseTree.Parse(line);
						if (tree != null) {
							Object result = tree.doEvaluate(locals, evaluator);
							lastRetVal = result;
							if (result == null) {
							} else {
								if (result.getClass() != Boolean.class) {
									writer.println(result.toString());
								}
							}
						} else {
							writer.println("Syntax error");
						}
					}
					exceptions = 0;
				} catch (Exception e) {
					writer.println("Error: " + e.getMessage());
					if (exceptions++ > 3) {
						writer.println("Too many exceptions");
						dontQuit = false;
					}
				}
			}
			dontQuit = false;
			writer.println("bye!");
			cleanup();
			if (ownsConnection) {
				try {
					System.out.println("Closing SQL connection...");
					connection.close();
				} catch (SQLException e) {
					e.printStackTrace();
				}
			}
		} catch (IOException e) {
			e.printStackTrace(writer);
		}
		dontQuit = false;
	}

	private void cleanup() {
		try {
			writer.println("Timeout; bye");
		} catch (Exception e) {
			e.printStackTrace();
		}
		try {
			reader.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
		try {
			input.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
		try {
			output.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
		try {
			writer.close();
		} catch (Exception e) {
			e.printStackTrace();
		}

	}

	public Object FunctionCallback(String funcName, Object[] vParams)
			throws WrongParameterCountException, FunctionNotFoundException {
		ClientFunctions function;

		if (calledFunctions++ > 128) {
			writer.println("Too many function calls!");
			this.dontQuit = false;
			return null;
		}

		try {
			function = ClientFunctions.valueOf(funcName);
		} catch (Exception e) {
			throw new FunctionNotFoundException(funcName);
		}

		if (this.isAdmin) {
			creationTime = Calendar.getInstance().getTimeInMillis() / 1000;
		}

		try {
			switch (function) {
			case reportflag:
				if (vParams.length != 2) {
					throw new WrongParameterCountException(funcName, 2,
							vParams.length);
				}
				return reportFlag((Integer) vParams[0], (String) vParams[1]);

			case help:
				if (vParams.length != 0) {
					throw new WrongParameterCountException(funcName, 0,
							vParams.length);
				}
				return Manual.getInstance();

			case quit:
				doQuit();
				return false;

			case log:
				if (this.isAdmin) {
					throw new FunctionNotFoundException(funcName);
				}
				writer.println("log: not allowed");
				return false;

			case list:
				if (this.isAdmin) {
					throw new FunctionNotFoundException(funcName);
				}
				writer.println("list: not allowed");
				return false;

			case array:
				if (this.isAdmin) {
					throw new FunctionNotFoundException(funcName);
				}
				writer.println("array: not allowed");
				return false;

			case dict:
				if (this.isAdmin) {
					throw new FunctionNotFoundException(funcName);
				}
				writer.println("dict: not allowed");
				return false;

			case readtext:
				if (vParams.length != 0) {
					throw new WrongParameterCountException(funcName, 0,
							vParams.length);
				}
				return readText();

			case reportadvisory:
				if (vParams.length != 2) {
					throw new WrongParameterCountException(funcName, 2,
							vParams.length);
				}
				return reportAdvisory((Integer) vParams[0], (String) vParams[1]);

			case range:
				if (this.isAdmin) {
					throw new FunctionNotFoundException(funcName);
				}
				writer.println("not allowed");
				return false;

			case admin:
				if (vParams.length > 1) {
					throw new WrongParameterCountException(funcName, 1,
							vParams.length);
				}
				if (this.isAdmin) {
					writer.println("Error: can't drop admin privileges!");
					this.isAdmin = false;
					this.evaluator
							.SetVariableMap(new HashMap<String, Object>());
					this.evaluator.setAllowArbitraryInstanciiation(false);
					this.evaluator.setTrusted(false);
					this.dontQuit = false;
					return true;
				}

				String enteredPassword;
				if (vParams.length == 0) {
					writer.write(Stuff.noLocalEcho());
					writer.print("Password: ");
					writer.flush();
					reader.munch(3);
					enteredPassword = reader.readLine();
					writer.write(Stuff.localEcho());
					reader.munch(3);
				} else {
					enteredPassword = (String) vParams[0];
				}

				if (enteredPassword.equals(SQLConnection.getInstance()
						.getProperty("adminPassword", "fooBAR#@!"))) {
					isAdmin = true;
					this.evaluator.SetVariable("db", SQLConnection
							.getInstance());
					this.evaluator.DeclareClassSafe(SQLConnection.class);
					writer.println("Login successful!");
					return true;
				}
				isAdmin = false;
				Thread.sleep(2000);
				writer.println("Login incorrect");

				if ((adminTriesLeft--) < 0) {
					this.dontQuit = false;
				}
				return false;

			case ladv:
				if (vParams.length != 0) {
					throw new WrongParameterCountException(funcName, 0,
							vParams.length);
				}
				if (!isAdmin) {
					writer.println("Requires admin privileges!");
					return false;
				}
				listAdvisories();
				return true;

			case functions:
				showFunctions();
				return true;

			case reject:
				if (!isAdmin) {
					writer.println("Requires admin privileges!");
					return false;
				}
				if (vParams.length != 2) {
					throw new WrongParameterCountException(funcName, 2,
							vParams.length);
				}
				markAdvisory((Integer) vParams[0], "rejected", 0,
						(String) vParams[1]);
				return true;
			case accept:
				if (!isAdmin) {
					writer.println("Requires admin privileges!");
					return false;
				}
				if (vParams.length != 3) {
					throw new WrongParameterCountException(funcName, 3,
							vParams.length);
				}
				markAdvisory((Integer) vParams[0], "accepted",
						(Integer) vParams[1], (String) vParams[2]);
				return true;

			case genflags:
				if (!isAdmin) {
					writer.println("Requires admin privileges!");
					return false;
				}
				if (vParams.length != 2) {
					throw new WrongParameterCountException(funcName, 2,
							vParams.length);
				}
				return genFlags((Integer) vParams[0], (String) vParams[1]);

			case delete:
				if (!isAdmin) {
					writer.println("Requires admin privileges!");
					return false;
				}
				if (vParams.length != 1) {
					throw new WrongParameterCountException(funcName, 3,
							vParams.length);
				}
				deleteAdvisory((Integer) vParams[0]);
				return true;

			case dechp:
				if (!isAdmin) {
					writer.println("Requires admin privileges!");
					return false;
				}
				if (vParams.length != 2) {
					throw new WrongParameterCountException(funcName, 2,
							vParams.length);
				}
				incHP((Integer) vParams[0], 0 - ((Integer) vParams[1]));
				return true;
			case inchp:
				if (!isAdmin) {
					writer.println("Requires admin privileges!");
					return false;
				}
				if (vParams.length != 2) {
					throw new WrongParameterCountException(funcName, 2,
							vParams.length);
				}
				incHP((Integer) vParams[0], (Integer) vParams[1]);
				return true;

			case createteam:
				if (!isAdmin) {
					writer.println("Requires admin privileges!");
					return false;
				}
				if (vParams.length != 1) {
					throw new WrongParameterCountException(funcName, 1,
							vParams.length);
				}
				return createTeam((String) vParams[0]);

			case getteam:
				if (!isAdmin) {
					writer.println("Requires admin privileges!");
					return false;
				}
				if (vParams.length != 1) {
					throw new WrongParameterCountException(funcName, 1,
							vParams.length);
				}
				return getTeam((Integer) vParams[0]);

			case createservice:
				if (!isAdmin) {
					writer.println("Requires admin privileges!");
					return false;
				}
				if (vParams.length != 1) {
					throw new WrongParameterCountException(funcName, 1,
							vParams.length);
				}
				return createService((String) vParams[0]);

			case getservice:
				if (!isAdmin) {
					writer.println("Requires admin privileges!");
					return false;
				}
				if (vParams.length != 1) {
					throw new WrongParameterCountException(funcName, 1,
							vParams.length);
				}
				return getService((Integer) vParams[0]);

			case listteams:
			case lt:
				return listTeams();

			case listservices:
			case ls:
				if (!isAdmin) {
					writer.println("Requires admin privileges!");
					return false;
				}
				return listServices();

			case getclass:
				if (!isAdmin) {
					writer.println("Requires admin privileges!");
					return false;
				}

				if (!SQLConnection.getInstance().getProperty("GetClass",
						"disabled").equals("enabled")) {
					writer
							.println("getclass is disabled. You can enable it in the 'settings' file");
					return false;
				}

				if (vParams.length != 1) {
					throw new WrongParameterCountException(funcName, 1,
							vParams.length);
				}
				this.evaluator.setAllowArbitraryInstanciiation(true);
				this.evaluator.setTrusted(true);
				Class cls = Class.forName((String) vParams[0]);
				if (cls != null) {
					this.evaluator.SetVariable(cls.getSimpleName(), cls);
					writer.println("Imported " + cls.getSimpleName());
					return true;
				} else {
					writer.println("Error: cannot import class "
							+ (String) vParams[0] + "!");
					return false;
				}

			case dir:
				for (int j = 0; j < 2; j++) {

					String[] keys = null;
					if (j == 0) {
						keys = this.evaluator.GetVariables();
					} else if (j == 1) {
						keys = new String[this.locals.keySet().size()];
						Iterator<String> itr = this.locals.keySet().iterator();
						int k = 0;
						while (itr.hasNext()) {
							keys[k++] = itr.next();
						}
					}
					for (int i = 0; i < keys.length; i++) {
						Object obj = null;
						if (j == 0) {
							obj = this.evaluator.GetVariable(keys[i]);
						} else if (j == 1) {
							obj = this.locals.get(keys[i]);
						}
						String desc = "null";
						if (obj != null) {
							if (obj != null) {
								desc = obj.toString();
							}
						}
						int pos = desc.indexOf("\n");
						if (pos != -1) {
							desc = desc.substring(0, pos - 1);
						}
						if (desc.length() > 40) {
							desc = desc.substring(0, 40) + "...";
						}
						writer.format("%s\t%s\t= %s\n", obj == null ? "" : obj
								.getClass().getSimpleName(), keys[i], desc);
					}
				}
				return true;
			}
		} catch (Exception e) {
			writer.println("Error: " + e.getMessage());
		}
		return null;
	}

	private boolean listServices() {
		try {
			ResultSet rs = connection
					.prepareStatement(
							"select service_name,service_script,service_check_interval,uid from services order by uid")
					.executeQuery();
			writer.println("ID\tname\tscript\tinterval");
			while (rs.next()) {
				writer.format("%d\t%s\t%s\t%s\n", rs.getInt(4),
						rs.getString(1), rs.getString(2), rs.getString(3));
			}
			return false;
		} catch (SQLException e) {
			e.printStackTrace(writer);
		}
		return false;
	}

	private ServiceManager createService(String serviceName) {
		try {
			PreparedStatement ps = connection
					.prepareStatement("insert into services(service_name,service_script_type) values(?,'binary')");
			ps.setString(1, serviceName);
			ps.execute();
			ps = connection
					.prepareStatement("select uid from services where service_name=?");
			ps.setString(1, serviceName);
			ResultSet rs = ps.executeQuery();
			if (!rs.next()) {
				writer.println("Error: couldn't create service " + serviceName
						+ "!");
				return null;
			}
			return new ServiceManager(connection, writer, rs.getInt(1));
		} catch (SQLException e) {
			e.printStackTrace();
			writer.println("Error: " + e.toString());
			return null;
		}
	}

	private ServiceManager getService(Integer serviceID) {
		try {
			PreparedStatement ps = connection
					.prepareStatement("select uid from services where uid=?");
			ps.setInt(1, serviceID);
			ResultSet rs = ps.executeQuery();
			if (!rs.next()) {
				writer.println("Error: service #" + serviceID
						+ " doesn't exist!");
				return null;
			}
			return new ServiceManager(connection, writer, serviceID);
		} catch (Exception e) {
			writer.println("Error: " + e.toString());
			return null;
		}
	}

	private TeamHandler getTeam(Integer teamID) {
		try {
			PreparedStatement ps = connection
					.prepareStatement("select uid from teams where uid=?");
			ps.setInt(1, teamID);
			ResultSet rs = ps.executeQuery();
			if (!rs.next()) {
				writer.println("Error: team #" + teamID + " doesn't exist!");
				return null;
			}
			return new TeamHandler(connection, writer, teamID);
		} catch (Exception e) {
			writer.println("Error: " + e.toString());
			return null;
		}
	}

	private boolean listTeams() {
		try {
			PreparedStatement ps = connection
					.prepareStatement("select team_points_offensive,team_points_defensive,team_points_advisories,team_points_hacking,team_name,uid,team_host from teams order by uid");
			ResultSet rs = ps.executeQuery();
			writer.println("ID\tTeam\toff\tdef\tadv\thak\thost");
			writer
					.println("-----------------------------------------------------------------");
			while (rs.next()) {
				String tName = rs.getString(5);
				String tHost = rs.getString(7);
				int tID = rs.getInt(6);
				int[] points = new int[4];
				if (this.isAdmin) {
					for (int i = 0; i < 4; i++) {
						points[i] = rs.getInt(i + 1);
					}
					writer.format("%d\t%s\t%d\t%d\t%d\t%d\t%s\n", tID, tName
							.substring(0, (tName.length() < 6) ? tName.length()
									: 6), points[0], points[1], points[2],
							points[3], tHost);
				} else {
					writer.format("%d\t%s\tn/a\tn/a\tn/a\tn/a\t%s\n", tID,
							tName.substring(0, (tName.length() < 6) ? tName
									.length() : 6), tHost);
				}
			}
			return true;
		} catch (SQLException e) {
			writer.println("Error: " + e);
		}
		return false;
	}

	private TeamHandler createTeam(String teamName) {
		try {
			PreparedStatement ps = connection
					.prepareStatement("insert into teams(team_name) values(?)");
			ps.setString(1, teamName);
			ps.execute();
			ps = connection
					.prepareStatement("select uid from teams where team_name=?");
			ps.setString(1, teamName);
			ResultSet rs = ps.executeQuery();
			if (!rs.next()) {
				writer.println("Error: couldn't create team " + teamName + "!");
				return null;
			}
			return new TeamHandler(connection, writer, rs.getInt(1));
		} catch (SQLException e) {
			e.printStackTrace();
			writer.println("Error: " + e.toString());
			return null;
		}
	}

	private void incHP(int teamID, int i) {
		PreparedStatement ps;
		ResultSet rs;

		try {
			ps = connection
					.prepareStatement("update teams set team_points_hacking=team_points_hacking+? where uid=?");
			ps.setInt(1, i);
			ps.setInt(2, teamID);
			ps.execute();
			ps = connection
					.prepareStatement("select team_points_hacking from teams where uid=?");
			ps.setInt(1, teamID);
			rs = ps.executeQuery();
			if (!rs.next()) {
				writer.println("Error: team " + teamID + " doesn't exist!");
			} else {
				writer.format("Team #%d now has %d hacking points\n", teamID,
						rs.getInt(1));
			}
		} catch (SQLException e) {
			e.printStackTrace(writer);
		}
	}

	private void deleteAdvisory(int advID) {
		PreparedStatement ps;
		ResultSet rs;

		try {
			ps = connection
					.prepareStatement("select advisory_description from advisories where uid=?");
			ps.setInt(1, advID);
			rs = ps.executeQuery();
			if (!rs.next()) {
				writer.println("Error: advisory #" + advID + " doesn't exist!");
			} else {
				ps = connection
						.prepareStatement("delete from advisories where uid=?");
				ps.setInt(1, advID);
				ps.execute();
				writer.println("Advisory deleted");
			}
		} catch (SQLException e) {
			e.printStackTrace(writer);
		}
	}

	private boolean genFlags(Integer numFlags, String separator) {
		try {
			int nfmo = numFlags - 1;
			for (int i = 0; i < numFlags; i++) {
				byte[] bytes = new byte[32];
				Random random = new Random();
				random.nextBytes(bytes);
				String flagID = Stuff.toHex(bytes);
				PreparedStatement ps = connection
						.prepareStatement("insert into flags(flag_name,flag_collected,flag_team,"
								+ "flag_service,flag_teamhost,flag_disttime) values (?,?,?,?,?,?)");
				ps.setString(1, flagID);
				ps.setBoolean(2, false);
				ps.setString(3, "WoD");
				ps.setString(4, "adlsmtp");
				ps.setString(5, "localhost");
				ps.setLong(6, Calendar.getInstance().getTimeInMillis() / 1000);
				ps.execute();
				// hack
				if (i == nfmo) {
					separator = "";
				}
				writer.print(flagID + separator);
			}
			writer.println();
			return true;
		} catch (SQLException e) {
			e.printStackTrace(writer);
		}
		return false;
	}

	private void markAdvisory(Integer advID, String status, int points,
			String note) {
		try {
			PreparedStatement ps = connection
					.prepareStatement("select advisory_team,advisory_status from advisories where uid=?");
			ps.setInt(1, advID);
			ResultSet rs = ps.executeQuery();
			if (!rs.next()) {
				writer.println("Advisory " + advID + " doesn't exist!");
				return;
			}
			if (!(rs.getString(2).equals("pending"))) {
				writer.println("Advisory " + advID + " was already rated!");
				return;
			}
			String teamName = rs.getString(1);
			connection.prepareStatement("begin").execute();
			ps = connection
					.prepareStatement("update advisories set advisory_status=? where uid=?");
			ps.setString(1, status);
			ps.setInt(2, advID);
			ps.execute();
			ps = connection
					.prepareStatement("update advisories set advisory_comment=? where uid=?");
			ps.setString(1, "[" + points + "] " + note);
			ps.setInt(2, advID);
			ps.execute();
			ps = connection
					.prepareStatement("select team_points_advisories from teams where team_name=?");
			ps.setString(1, teamName);
			rs = ps.executeQuery();
			rs.next();
			ps = connection
					.prepareStatement("update teams set team_points_advisories=? where team_name=?");
			ps.setInt(1, rs.getInt(1) + points);
			ps.setString(2, teamName);
			ps.execute();
			connection.prepareStatement("commit").execute();
			writer.println("ok");
		} catch (Exception e) {
			e.printStackTrace(writer);
		}
	}

	private void listAdvisories() {
		try {
			PreparedStatement ps = connection
					.prepareStatement("select uid,advisory_team,advisory_description from advisories where advisory_status='pending' order by advisory_time");
			ResultSet rs = ps.executeQuery();
			while (rs.next()) {
				String desc = rs.getString(3);
				if (desc == null) {
					desc = "";
				}
				writer.println(rs.getLong(1) + ": [" + rs.getString(2) + "]  "
						+ desc.replace('\n', ' '));
			}
		} catch (Exception e) {
			e.printStackTrace(writer);
		}
	}

	private boolean reportAdvisory(Integer teamID, String advisory) {
		try {
			String teamName;
			PreparedStatement ps = connection
					.prepareStatement("select team_name from teams where uid=?");
			ps.setInt(1, teamID);
			ResultSet rs = ps.executeQuery();
			if (!rs.next()) {
				writer.println("Team # " + teamID + " does not exist!");
				return false;
			}
			teamName = rs.getString(1);
			ps = connection
					.prepareStatement("insert into advisories (advisory_team,advisory_description,advisory_time,advisory_status,advisory_from) values (?,?,?,?,?)");
			ps.setString(1, teamName);
			ps.setString(2, advisory);
			ps.setLong(3, Calendar.getInstance().getTimeInMillis() / 1000);
			ps.setString(4, "pending");
			ps.setString(5, this.socket.getInetAddress().toString());
			ps.execute();
			writer.println("Your advisory has been reported.");
			return true;
		} catch (SQLException e) {
		}

		return false;
	}

	private Object readText() {
		String string = "";
		int lines = 0;

		writer
				.println("Write some text, finish with a single dot on a separate line (^\\.$)");

		String line;
		while (true) {
			try {
				writer.print(lines + "> ");
				writer.flush();
				line = reader.readLine();
			} catch (IOException e) {
				return e;
			}
			if (line.equals(".")) {
				break;
			}
			string = string + line + "\n";
			if (lines++ > 100) {
				break;
			}
		}

		return string;
	}

	private void doQuit() {
		dontQuit = false;
	}

	private void showFunctions() {
		ClientFunctions[] functions = ClientFunctions.values();
		writer.println("Available functions");
		writer
				.println("-----------------------------------------------------------------");
		for (int i = 0; i < functions.length; i++) {
			writer.println(functions[i].toString());
		}
	}

	private boolean reportFlag(Integer integer, String string) {
		if (reportedFlags++ > 100) {
			writer.println("Too many flags!");
			this.dontQuit = false;
			return false;
		}
		try {
			PreparedStatement ps = connection
					.prepareStatement("select * from flags where flag_name=?");
			ps.setString(1, string);
			ResultSet rs = ps.executeQuery();
			if (!rs.next()) {
				writer.println("Flag \"" + string + "\" does not exist!");
			} else {
				if (rs.getBoolean(3)) {
					writer.println("This flag is not valid anymore!");
				} else {
					PreparedStatement ps2;
					ps2 = connection
							.prepareStatement("select team_name,team_points_offensive from teams where uid=?");
					ps2.setInt(1, integer);
					ResultSet rs2 = ps2.executeQuery();
					if (!rs2.next()) {
						writer.println("Team ID " + integer + " invalid!");
						return false;
					}
					String teamName = rs2.getString(1);
					if (teamName.equals(rs.getString(5))) {
						writer.println("You cannot report your own flags!");
						return false;
					}
					int points = rs2.getInt(2);
					writer.println("You successfully reported flag " + string
							+ ", which is from service " + rs.getString(6));
					ps2 = connection
							.prepareStatement("update flags set flag_collected=true where uid="
									+ rs.getBigDecimal(1));
					ps2.execute();
					ps2 = connection
							.prepareStatement("update flags set flag_collectingteam=? where uid="
									+ rs.getBigDecimal(1));
					ps2.setString(1, teamName);
					ps2.execute();
					ps2 = connection
							.prepareStatement("update teams set team_points_offensive=team_points_offensive+1 where uid=?");
					ps2.setInt(1, integer);
					ps2.execute();
					writer.println("You now have " + (points + 1)
							+ " offensive points");
					return true;
				}
			}
		} catch (SQLException e) {
			writer.println("Error: " + e.getLocalizedMessage());
		}
		return false;
	}

	public void doStop() {
		dontQuit = false;
	}

	public ComparableBA getIP() {
		return new ComparableBA(ipAddress);
	}

	public boolean isQuit() {
		return !dontQuit;
	}

	@Override
	public String toString() {
		return this.socket.getInetAddress().toString();
	}
}
